/*
 * Copyright (C) 2018 Gravity Engineering Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.graveng.rci;

import java.beans.XMLDecoder;
import java.beans.XMLEncoder;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Base64;
import java.util.HashMap;

import com.graveng.rci.util.ClassField;

/**
 * RCIサーバとRCIクライアント間でxml形式で送受信されるパケット。
 * <p>
 * このクラスのpublicメソッドの多くは、RciとRciServerに使われることを想定しており、
 * クライアントユーザが直接呼び出すのは返却値を受け取るgetResult, toResult程度である。<br>
 * <br>
 * 遠隔メソッドコール時のパラメータ領域のクラスがリモートで変更された場合もクライアントで反映される。<br>
 * <br>
 * publicにしている大きな理由は、パッケージ間のアクセスがjavaではそうせざるを得ないということだけではなく、
 * パケットを転送する際にオブジェクトをxmlにするためのXMLEncoderがpublicのsetter, getterを必要とするからである。
 * (そのためどこからも呼ばれないようなsetterも削除してはいけない。)
 */
public class Packet {

	/**
	 * サーバで発生したイベントをクライアントで受信してリスナーを起動する際に識別する番号
	 */
	public static final String OPTION_LISTENER_NUMBER = "listerNumber";

	/**
	 * サーバからイベント通知であることを示すオプション
	 */
	public static final String OPTION_CALLBACK = "callback";

	/**
	 * サーバでイベントを受け取るクラスのハッシュコード
	 */
	public static final String OPTION_LISTENER_HASHCODE = "listerHashcode";

	/**
	 * パラメータの型を登録するオプション
	 */
	public static final String OPTION_PARAMTYPES = "paramTypes";

	/**
	 * 要求に対する応答であることを識別する情報（特に使われていない）
	 */
	protected long requestId = System.currentTimeMillis();

	/**
	 * メソッド名を含む可変パラメータ
	 */
	protected Object[] params;

	/**
	 * パラメータをシリアライズして転送するための領域。
	 */
	protected String[] serializedParams;

	/**
	 * サーバ側でクラスのインスタンスを識別する番号
	 */
	protected int instanceNumber = 0;

//遠隔生成するクラスのインスタンスの転送は行うべきではないであろうから、コメントアウトしている。
// (将来的にオブジェクトを)
//	protected Object instance;

	/**
	 * メソッドの返却値の相当するObject
	 */
	protected Object result;

	/**
	 * サーバで発生した例外
	 */
	protected Exception exception;

	/**
	 * サーバで発生した例外をシリアライズする領域
	 */
	protected String serializedException;

	/**
	 * 各種オプションを保持する領域
	 */
	protected HashMap<Object, Object> optionMap = new HashMap<Object, Object>();

	/**
	 * デフォルトコンストラクタ
	 */
	public Packet() {

	}

	/**
	 * 要求電文と応答電文を関連づけるリクエストIDを取得する。
	 * @return リクエストID
	 */
	public long getRequestId() {
		return requestId;
	}

	/**
	 * 要求電文と応答電文を関連づけるリクエストIDを設定する。
	 * @param requestId リクエストID
	 */
	public void setRequestId(long requestId) {
		this.requestId = requestId;
	}

	/**
	 * シリアライズされたパラメータを取得する。
	 * シリアライズしていない対応するパラメータはnullになっている。
	 * @return シリアライズされたパラメータ
	 */
	public String[] getSerializedParams() {
		return serializedParams;
	}

	/**
	 * シリアライズされたパラメータを設定する。
	 * @param serializedParams シリアライズされたパラメータ
	 */
	public void setSerializedParams(String[] serializedParams) {
		this.serializedParams = serializedParams;
	}

	/**
	 * オプションマップを取得する。
	 * @return オプションマップ
	 */
	public HashMap<Object, Object> getOptionMap() {
		return optionMap;
	}

	/**
	 * オプションマップを設定する。
	 * @param optionMap オプションマップ
	 */
	public void setOptionMap(HashMap<Object, Object> optionMap) {
		this.optionMap = optionMap;
	}

	/**
	 * パラメータを取得する。
	 * @return パラメータ
	 */
	public Object[] getParams() {
		return params;
	}

	/**
	 * パラメータを設定する。
	 * @param params パラメータ
	 */
	public void setParams(Object[] params) {
		this.params = params;
		for(int i = 0; i < this.params.length; i++) {
			if(this.params[i] == null) {
				this.params[i] = new NullObject();
			}
		}
	}

	/**
	 * インスタンス番号を取得する。
	 * @return インスタンス番号
	 */
	public int getInstanceNumber() {
		return instanceNumber;
	}

	/**
	 * インスタンス番号を設定する。
	 * @param instanceNumber インスタンス番号
	 */
	public void setInstanceNumber(int instanceNumber) {
		this.instanceNumber = instanceNumber;
	}

	/**
	 * サーバで実行したメソッドの返却値相当を受け取る
	 * @return 返却値
	 */
	public Object getResult() {
		return result;
	}

	@SuppressWarnings("unchecked")
	/**
	 * サーバで実行したメソッドの返却値相当を型変換して受け取る
	 * @param to 返却値がStringなら""を与えればいいし、intなら0を与えて変換する型指定する。
	 * @return 返却値
	 */
	public <T> T toResult(T to) {
		return (T) result;
	}

	/**
	 * 遠隔メソッドコールの返却値を設定する。
	 * @param result 返却値
	 */
	public void setResult(Object result) {
		this.result = result;
	}

	/**
	 * サーバで発生した例外を取得する。
	 * @return 例外
	 */
	public Exception getException() {
		return exception;
	}

	/**
	 * サーバで発生した例外を設定する。
	 * @param exception 例外
	 */
	public void setException(Exception exception) {
		this.exception = exception;
	}

	/**
	 * シリアライズされたサーバで発生した例外を取得する。
	 * @return シリアライズされたBase64例外
	 */
	public String getSerializedException() {
		return serializedException;
	}

	/**
	 * シリアライズされたサーバで発生した例外を設定する。
	 * @param serializedException シリアライズされたBase64例外
	 */
	public void setSerializedException(String serializedException) {
		this.serializedException = serializedException;
	}

//遠隔生成するクラスのインスタンスの転送は行うべきではないので、コメントアウトしているが、将来用途もあるかも知れない
// ので一応残しておく。
//	/**
//	 * @return instance
//	 */
//	public Object getInstance() {
//		return instance;
//	}
//
//	@SuppressWarnings("unchecked")
//	public <T> T toInstance(T to) {
//		return (T) instance;
//	}
//
//	/**
//	 * @param instance
//	 *            セットする instance
//	 */
//	public void setInstance(Object instance) {
//		this.instance = instance;
//	}

	/**
	 * パラメータをシリアライズする。
	 * @throws IOException IO例外
	 */
	public void enSerializeParam() throws IOException {
		serializedParams = new String[params.length];
		for (int i = 0; i < params.length; i++) {
			if (params[i] instanceof Serializable) {
				if (params[i] instanceof String) {
					serializedParams[i] = null;
				} else if (isPrimitive(params[i])) {
					serializedParams[i] = null;
				} else {
					serializedParams[i] = getEncodeStr(params[i]);
				}
			} else {
				serializedParams[i] = null;
			}
		}

		if(exception != null) {
			serializedException = getEncodeStr(exception);
			exception = null;
		}
	}

	/**
	 * パラメータをデシリアライズする。
	 * @throws Exception 何らかの例外
	 */
	public void deSerializeParam() throws Exception {
		for (int i = 0; i < params.length; i++) {
//			if (params[i] == null && serializedParams[i] != null) {
			if ((params[i] == null || params[i] instanceof NullObject) && serializedParams[i] != null) {
				params[i] = getDecodeStr(serializedParams[i]);
			}
		}
		if(serializedException != null) {
			exception = (Exception)getDecodeStr(serializedException);
		}
	}

// 以下にはユーティリティのようなものが並んでいる

	/**
	 * ObjectをXML形式の文字列として返す。
	 *
	 * @param obj Object
	 * @return Objectのxml形式
	 * @throws IOException IO例外
	 */
	public static String getEncodeXML(Object obj) throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		XMLEncoder encoder = new XMLEncoder(baos);
		encoder.writeObject(obj);
		encoder.close();
		String result = baos.toString();
		return result;
	}

	/**
	 * XML形式の文字列からObjectを生成する。
	 *
	 * @param xmlObject Objectのxml形式
	 * @return Object
	 * @throws IOException IO例外
	 */
	public static Object getDecodeXML(String xmlObject) throws IOException {
		ByteArrayInputStream baos = new ByteArrayInputStream(xmlObject.getBytes());
		XMLDecoder decoder = new XMLDecoder(baos);
		Object obj = decoder.readObject();
		decoder.close();
		return obj;
	}

	/**
	 * オブジェクトをシリアライズしBase64にして返す。
	 * @param obj Obejct
	 * @return ObjectシリアライズのBase64形式
	 * @throws IOException IO例外
	 */
	public static String getEncodeStr(Object obj) throws IOException {
		return enDumpByte(getEncode(obj));
	}

	/**
	 * オブジェクトをシリアライズする。
	 * @param obj Obejct
	 * @return Objectシリアライズ
	 * @throws IOException IO例外
	 */
	public static byte[] getEncode(Object obj) throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ObjectOutputStream oos = new ObjectOutputStream(baos);
		oos.writeObject(obj);
		oos.close();
		return baos.toByteArray();
	}

	public static Object getDecodeStr(String obj) throws Exception {
		return getDecode(deDumpByte(obj));
	}

	/**
	 * オブジェクトをデシリアライズする。
	 * @param buf Objectシリアライズ
	 * @return Obejct
	 * @throws IOException IO例外
	 * @throws ClassNotFoundException クラスが見つからない
	 */
	public static Object getDecode(byte[] buf) throws IOException, ClassNotFoundException {
		ObjectInputStream ois = new ObjectInputStream(new ByteArrayInputStream(buf));
		Object o = ois.readObject();
		ois.close();
		return o;
	}

	/**
	 * バイナリの文字列(Base64)形式を返す
	 * @param buf バイナリ
	 * @return 文字列(Base64)形式
	 */
	public static String enDumpByte(byte[] buf) {
//		return enHexByte(buf);
		return enBase64Byte(buf);
	}

	/**
	 * バイナリの文字列(16進数)形式を返す
	 * @param buf バイナリ
	 * @return 文字列(16進数)形式
	 */
	public static String enHexByte(byte[] buf) {
		// １６進数文字列に変換
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < buf.length; i++) {
			int intValue = 0;
			intValue = 0xFF & buf[i];
			String text = String.format("%02x", intValue);
			sb.append(text);
		}
		return sb.toString();
	}

	/**
	 * バイナリの文字列(Base64)形式を返す
	 * @param buf バイナリ
	 * @return 文字列(Base64)形式
	 */
	public static String enBase64Byte(byte[] buf) {
		// Base64に変換
		return new String(Base64.getEncoder().encode(buf));
	}

	/**
	 * Base64形式をバイナリに変換する。
	 * @param text Base64形式
	 * @return バイナリ
	 */
	public static byte[] deDumpByte(String text) {
//		return deHexByte(text);
		return deBase64Byte(text);
	}

	/**
	 * 16進数文字列形式をバイナリに変換する。
	 * @param text 16進数文字列
	 * @return バイナリ
	 */
	public static byte[] deHexByte(String text) {
		byte[] buf = text.getBytes();
		byte[] result = new byte[buf.length / 2];

		for (int i = 0; i < result.length; i++) {
			String sub = text.substring(i * 2, (i * 2) + 2);
			int byteValue = Integer.parseInt(sub, 16) & 0xFF;
			result[i] = (byte) (0xFF & byteValue);
		}
		return result;
	}

	/**
	 * Base64文字列形式をバイナリに変換する。
	 * @param text Base64形式文字列
	 * @return バイナリ
	 */
	public static byte[] deBase64Byte(String text) {
		return Base64.getDecoder().decode(text.getBytes());
	}

	/**
	 * 実行しているメソッド名を取得する。
	 * @return メソッド名
	 */
	public static String getMethodName() {
		Throwable t = new Throwable();
		// String className = t.getStackTrace()[1].getClassName();
		String methodName = t.getStackTrace()[1].getMethodName();
		return methodName;
	}

	/**
	 * オブジェクトをコピーする。
	 * @param from コピー元Object
	 * @param to コピー先Object
	 * @return コピーに成功したらtrue
	 * @throws IllegalArgumentException パラメータエラー
	 * @throws IllegalAccessException 不正アクセス
	 */
	public static boolean copyObject(Object from, Object to) throws IllegalArgumentException, IllegalAccessException {
		return ClassField.clone(from, to);
	}

	/**
	 * オブジェクトの型を求める
	 * @param obj Object
	 * @return Objectの型
	 */
	public static Class<?> convertPrimitive(Object obj) {
		if (obj == null) {
			return null;
		}
		// 実際にInteger型がパラメータなら見つけられないことになるが、普通はそうしないからこれでほとんど問題ない。
		if (obj instanceof java.lang.Integer) {
			return int.class;
		}
		if (obj instanceof java.lang.Long) {
			return long.class;
		}
		if (obj instanceof java.lang.Double) {
			return double.class;
		}
		if (obj instanceof java.lang.Boolean) {
			return boolean.class;
		}
		if (obj instanceof java.lang.Short) {
			return short.class;
		}
		if (obj instanceof java.lang.Float) {
			return float.class;
		}
		if (obj instanceof java.lang.Byte) {
			return byte.class;
		}
		if (obj instanceof java.lang.Character) {
			return char.class;
		}
		return obj.getClass();
	}

	/**
	 * オブジェクトのラッパークラスがプリミティブ型ならtrueを返す。
	 * @param obj オブジェクト
	 * @return ラッパークラスがプリミティブ型ならtrue
	 */
	public static Boolean isPrimitive(Object obj) {
		if (obj == null) {
			return false;
		}
		if (obj instanceof java.lang.Integer) {
			return true;
		}
		if (obj instanceof java.lang.Long) {
			return true;
		}
		if (obj instanceof java.lang.Double) {
			return true;
		}
		if (obj instanceof java.lang.Boolean) {
			return true;
		}
		if (obj instanceof java.lang.Short) {
			return true;
		}
		if (obj instanceof java.lang.Float) {
			return true;
		}
		if (obj instanceof java.lang.Byte) {
			return true;
		}
		if (obj instanceof java.lang.Character) {
			return true;
		}
		return false;
	}

	// ArrayList<String> result = (ArrayList<String>)packet.getResult();
	//
	// 上を以下にすれば、warningが出なくなる。
	//
	// ArrayList<String> result = null;
	// result = Packet.cast(result, packet.getResult());
	//
	@SuppressWarnings("unchecked")
	/**
	 * オブジェクトのキャスト。
	 * @param to キャスト後
	 * @param from キャスト前
	 * @return キャスト後
	 */
	public static <T> T cast(T to, Object from) {
		return (T) from;
	}

}

