/*
 * Copyright (C) 2018 Gravity Engineering Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.graveng.rci.crypt;

import java.util.Base64;

import javax.crypto.Cipher;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;

import org.apache.commons.lang3.RandomStringUtils;

import com.graveng.rci.util.FileUtil;

/**
 * AesCbc方式の共通鍵による暗号化を行う。
 * <p>
 * AES(Advanced Encryption Standard) CBC(Cipher Block Chaining mode) 128(キー長、128ビット)<br>
 * Aesのキー長は256ビットが望ましいが、javaの標準ライブラリの輸出規制の制約から標準では128ビットにしている。
 * 256ビットに変更するには、${JAVA_HOME}/jre/lib/securityのlocal_policy.jarとUS_export_policy.jarを
 * 無制限のものに置き換え、com.graveng.rci.crypt.AesCrypt.keyLengthを16から32(256ビット)に変更する。<br>
 * <br>
 * キー長が128ビットでも暗号強度を格段に高めるために暗号化／復号化を繰り返すことができる。<br>
 * ServerConfig.xmlでその繰り返し回数を変更できる。（デフォルトは1）
 */
public class AesCrypt {
	/**
	 * テキスト変換時のエンコーディング
	 */
	protected String encoding = System.getProperty("file.encoding");

	/**
	 * キー長が、128ビットなら16(デフォルト)を設定。
	 */
	public static int keyLength = 16;
//	protected static int keyLength = 32;

	/**
	 * 暗号化用ライブラリ
	 */
    protected Cipher encrypter = null;

    /**
     * 復号化用ライブラリ
     */
    protected Cipher decrypter = null;

    /**
     * 自動生成される共通鍵の初期値
     */
    protected byte[] secretKey = RandomStringUtils.randomAlphanumeric(keyLength).getBytes();

    /**
     * 自動生成されるイニシャルベクター(暗号化時のスタートブロック用の初期値 128ビット固定長）
     */
    protected byte[] ivs = RandomStringUtils.randomAlphanumeric(16).getBytes();

    /**
     * アルゴリズム名
     */
//    public static String Algorithm = "AES/CBC/PKCS5Padding";
    public static String Algorithm = "AES/CBC/NOPADDING";

    /**
     * デフォルトコンストラクタ
     */
    public AesCrypt() {
//    	int length = Cipher.getMaxAllowedKeyLength("AES");
//    	System.out.println(length);
    }

    /**
     * エンコーディングを指定してAesCryptを生成
     * @param encoding エンコーディング、UTF-8等
     */
    public AesCrypt(String encoding) {
    	setEncoding(encoding);
    }

    /* コンストラクタ（引数は暗号解読キーとIV）
     *   引数は、どちらもStringからbyte[]に変換したものを指定する
     */
    public AesCrypt(byte[] secretKey, byte[] ivs) throws Exception{
    	this.secretKey = secretKey;
    	this.ivs = ivs;
    }

	/**
     * 暗号化ライブラリの初期化
	 * @throws Exception 何らかの例外
     */
    protected void init() throws Exception {
    	// 暗号化時のスタートブロック用の初期値を作成
        IvParameterSpec iv = new IvParameterSpec(ivs);

        SecretKeySpec key = new SecretKeySpec(secretKey, "AES");
        encrypter = Cipher.getInstance(Algorithm);
        encrypter.init(Cipher.ENCRYPT_MODE, key, iv);

        decrypter = Cipher.getInstance(Algorithm);
        decrypter.init(Cipher.DECRYPT_MODE, key, iv);
    }

    /**
     * String変換する際のエンコーディングを取得する。
	 * @return encoding エンコーディング、UTF-8等
	 */
	public String getEncoding() {
		return encoding;
	}

	/**
	 * String変換する際のエンコーディングを設定する。
	 * @param encoding エンコーディング、UTF-8等
	 */
	public void setEncoding(String encoding) {
		this.encoding = encoding;
	}

	/**
     * 共通鍵を取得する。
     * @return 共通鍵
     */
	public byte[] getSecretKey() {
		return secretKey;
	}

	/**
	 * 共通鍵を設定する。
	 * @param secretKey 共通鍵
	 */
	public void setSecretKey(byte[] secretKey) {
		// 16バイト分までコピーするため
		for(int i = 0; i < secretKey.length && i < this.secretKey.length; i++) {
			this.secretKey[i] = secretKey[i];
		}
		encrypter = null;
		decrypter = null;
	}

	/**
	 * 初期ベクターを取得する。
	 * @return 初期ベクター
	 */
	public byte[] getIvs() {
		return ivs;
	}

	/**
	 * 初期ベクターを設定する。
	 * @param ivs 初期ベクター
	 */
	public void setIvs(byte[] ivs) {
		// 16バイト分までコピーするため
		for(int i = 0; i < ivs.length && i < this.ivs.length; i++) {
			this.ivs[i] = ivs[i];
		}
		encrypter = null;
		decrypter = null;
	}

	/**
	 * 暗号化する。
	 * @param bytes 平文
	 * @return 暗号文
	 * @throws Exception 何らかの例外
	 */
    public byte[] encrypt(byte[] bytes) throws Exception {
    	return encrypt(bytes, 1);
    }

    /**
	 * 暗号化する。
	 * @param bytes 平文
	 * @param repeat 繰り返し数
	 * @return 暗号文
	 * @throws Exception 何らかの例外
	 */
    public byte[] encrypt(byte[] bytes, int repeat) throws Exception {
    	if(encrypter == null) {
    		init();
    	}
//    	byte[] result = bytes;
    	byte[] result = userSpacePadding(bytes);
    	for(int i = 0; i < repeat; i++) {
    		result = encrypter.doFinal(result);
    	}
    	return result;
    }

    /**
     * 入力データが16の倍数ではない場合にスペースパディングする
     * @param in 入力データ
     * @return パディングされたデータ
     */
    protected byte[] userSpacePadding(byte[] in) {
    	byte[] out = in;
    	if(Algorithm.toUpperCase().indexOf("NOPADDING") > 0) {
    		int amari = in.length % 16;
    		if(amari != 0) {
    			out = new byte[in.length + (16-amari)];
    			System.arraycopy(in, 0,out, 0, in.length);
    			for(int i = in.length; i < out.length; i++ ) {
    				out[i] = 0x20;
    			}
    		}
    	}
    	return out;
    }

    /**
     * 復号化する。
     * @param bytes 暗号文
     * @return 平文
     * @throws Exception 何らかの例外
     */
    public byte[] decrypt(byte[] bytes) throws Exception {
    	return decrypt(bytes, 1);
    }

    /**
     * 復号化する。
     * @param bytes 暗号文
     * @param repeat 繰り返し回数
     * @return 平文
     * @throws Exception 何らかの例外
     */
    public byte[] decrypt(byte[] bytes, int repeat) throws Exception {
    	if(decrypter == null) {
    		init();
    	}
    	byte[] result = bytes;
    	for(int i = 0; i < repeat; i++) {
    		result = decrypter.doFinal(result);
    	}
    	return result;
    }

	/**
	 * 暗号化し、さらにBase64にする。
	 * @param text 平文
	 * @return Base64暗号文
	 * @throws Exception 何らかの例外
	 */
    public String encrypt(String text) throws Exception {
    	return encrypt(text, 1); // １回の暗号化
    }

	/**
	 * 暗号化し、さらにBase64にする。
	 * @param text 平文
	 * @param repeat 繰り返し回数
	 * @return Base64暗号文
	 * @throws Exception 何らかの例外
	 */
    public String encrypt(String text, int repeat) throws Exception {
        byte[] bytes = text.getBytes(encoding);
        for(int i = 0; i < repeat; i++) {
        	bytes = encrypt(bytes);
        }
        // 暗号化したものをBase64で文字列化する
        byte[] str64 = Base64.getEncoder().encode(bytes);
        return new String(str64, encoding);
    }

    /**
	 * 暗号化し、さらにBase64にする。
	 * @param bytes 平文
	 * @return Base64暗号文
	 * @throws Exception 何らかの例外
	 */
    public String encryptToBase64(byte[] bytes) throws Exception {
        byte[] crypt = encrypt(bytes);
        // 暗号化したものをBase64で文字列化する
        byte[] str64 = Base64.getEncoder().encode(crypt);
        return new String(str64, encoding);
    }

    /**
     * Base64の暗号文を復号化する。
     * @param str64 Base64暗号文
     * @return 平文
     * @throws Exception 何らかの例外
     */
    public String decrypt(String str64) throws Exception {
    	return decrypt(str64, 1); // 1回の復号化
    }

    /**
     * Base64の暗号文を復号化する。
     * @param str64 Base64暗号文
     * @param repeat 繰り返し回数
     * @return 平文
     * @throws Exception 何らかの例外
     */
    public String decrypt(String str64, int repeat) throws Exception {
    	// Base64で文字列化されていたものをバイナリに戻す
    	byte[] bytes = Base64.getDecoder().decode(str64.getBytes(encoding));
    	for(int i = 0; i < repeat; i++) {
    		bytes = decrypt(bytes);
    	}
        // 文字列にして返す
        return new String(bytes, encoding);
    }

    /**
     * Base64の暗号文を復号化する。
     * @param str64 Base64暗号文
     * @return 平文
     * @throws Exception 何らかの例外
     */
    public byte[] decryptFromBase64(String str64) throws Exception {
    	// Base64で文字列化されていたものをバイナリに戻す
    	byte[] str = Base64.getDecoder().decode(str64.getBytes(encoding));
        byte[] text = decrypt(str);
        return text;
    }

    /**
     * ファイルを暗号化／復号化する。(単純にファイルのすべてをメモリに読み込んで処理するので巨大なファイルは無理)
     * <p>
     * @param args
     * [0]:暗号化か復号化か指定 "-e" か "-d"、"-eb", "-bd"なら暗号化しBase64化<br>
     * [1]:共通鍵(最大16バイトの文字列)<br>
     * [2]:初期ベクター(最大16バイトの文字列)<br>
     * [3]:入力ファイルパス<br>
     * [4]:出力ファイルパス<br>
     */
	public static void main(String[] args) {
		String usage = "usage:AesCrypt -e|-d|-eb|-bd COMMON_KEY INITIAL_VECTOR IN_FILE_PATH OUT_FILE_PATH";
		if(args.length != 5) {
			System.out.println(usage);
			System.exit(1);
		}

		AesCrypt.Algorithm = "AES/CBC/PKCS5Padding";
		int repeat = 1;

		byte[] secretKey = new byte[32];
		byte[] iniVec = new byte[16];

		try {
			AesCrypt crypt = new AesCrypt(secretKey, iniVec);
			crypt.setSecretKey(args[1].getBytes());
			crypt.setIvs(args[2].getBytes());

			if(args[0].equals("-e")) {
				// 暗号化を行う
				byte[] inData = FileUtil.loadFile(args[3]);
				byte[] outData = crypt.encrypt(inData, repeat);
				FileUtil.saveFile(args[4], outData);
			}
			else if(args[0].equals("-d")) {
				// 復号化を行う
				byte[] inData = FileUtil.loadFile(args[3]);
				byte[] outData = crypt.decrypt(inData, repeat);
				FileUtil.saveFile(args[4], outData);
			}
			else if(args[0].equals("-eb")) {
				// 暗号化を行う
				byte[] inData = FileUtil.loadFile(args[3]);
				String outData = crypt.encryptToBase64(inData);
				FileUtil.saveFile(args[4], outData.getBytes());
			}
			else if(args[0].equals("-bd")) {
				// 復号化を行う
				byte[] inData = FileUtil.loadFile(args[3]);
				byte[] outData = crypt.decryptFromBase64(new String(inData));
				FileUtil.saveFile(args[4], outData);
			}
			else {
				System.out.println(usage);
				System.exit(1);
			}

		} catch (Exception e) {
			System.out.println(usage);
			e.printStackTrace();
			System.exit(3);
		}
		System.exit(0);
	}

}
