/*
 * Copyright (C) 2018 Gravity Engineering Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.graveng.rci.util;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;

/**
 * リフレクションを利用したクラスの生成やコピー、変数表示を行う。
 */
public class ClassField {

	private ClassField() {
	}

	/**
	 * 親クラスを含めたオブジェクトの各種変数を文字列形式で取得する
	 * @param obj オブジェクト
	 * @return オブジェクトの各種変数の文字列形式
	 */
	public static String toDump(Object obj) {
		return toDumpSuper(obj);
	}

	/**
	 * 自オブジェクトの各種変数を文字列形式で取得する
	 * @param obj オブジェクト
	 * @return オブジェクトの各種変数の文字列形式
	 */
	public static String toDumpOwn(Object obj) {
		StringBuilder sb = new StringBuilder();
		try {
			Class<? extends Object> c = obj.getClass();
			Field[] f =  c.getDeclaredFields();
			for(int i = 0; i < f.length; i++) {
				f[i].setAccessible(true);
				if(Modifier.isTransient(f[i].getModifiers())) {
					continue;
				}
				if (!Modifier.isStatic(f[i].getModifiers())) {
					sb.append("[");
					sb.append(f[i].getName());
					sb.append("=");
					if(f[i].get(obj) != null) {
						sb.append(f[i].get(obj).toString());
					}
					sb.append("] ");
				}
			}
		}catch(Exception e) {
			Logger.warn("fail", e);
		}
		return sb.toString();
	}

	/**
	 * 親クラスを含めたオブジェクトの各種変数を文字列形式で取得する
	 * @param obj オブジェクト
	 * @return オブジェクトの各種変数の文字列形式
	 */
	public static String toDumpSuper(Object obj) {
		StringBuilder sb = new StringBuilder();
		try {
			Class<? extends Object> c = obj.getClass();

			Field[] f =  c.getDeclaredFields();
			for(int i = 0; i < f.length; i++) {
				f[i].setAccessible(true);
				if(Modifier.isTransient(f[i].getModifiers())) {
					continue;
				}
				if (!Modifier.isStatic(f[i].getModifiers())) {
					sb.append("[");
					sb.append(f[i].getName());
					sb.append("=");
					if(f[i].get(obj) != null) {
						sb.append(f[i].get(obj).toString());
					}
					sb.append("] ");
				}
			}

			// 続いてsuperクラス
			c = c.getSuperclass();
			f =  c.getDeclaredFields();
			for(int i = 0; i < f.length; i++) {
				f[i].setAccessible(true);
				if(Modifier.isTransient(f[i].getModifiers())) {
					continue;
				}
				if (!Modifier.isStatic(f[i].getModifiers())) {
					sb.append("[");
					sb.append(f[i].getName());
					sb.append("=");
					if(f[i].get(obj) != null) {
						sb.append(f[i].get(obj).toString());
					}
					sb.append("] ");
				}
			}

		}catch(Exception e) {
			Logger.warn("fail", e);
		}
		return sb.toString();
	}

	/**
	 * オブジェクトを複製する
	 * @param from コピー元オブジェクト
	 * @param to コピー先オブジェクト
	 * @return コピーに成功したらtrue
	 * @throws IllegalArgumentException パラメータエラー
	 * @throws IllegalAccessException 不正なアクセス
	 */
	public static boolean clone(Object from, Object to) throws IllegalArgumentException, IllegalAccessException {
		if(!from.getClass().getName().equals(to.getClass().getName())) {
			return false;
		}

		Class<? extends Object> cfrom = from.getClass();
		Class<? extends Object> cto = to.getClass();

		Field[] ffrom =  cfrom.getDeclaredFields();
		Field[] fto =  cto.getDeclaredFields();
		for(int i = 0; i < ffrom.length; i++) {
			ffrom[i].setAccessible(true);
			fto[i].setAccessible(true);
			if(Modifier.isTransient(ffrom[i].getModifiers())) {
				continue;
			}
			if (!Modifier.isStatic(ffrom[i].getModifiers())) {
				String key = null,  type = null;
				key = ffrom[i].getName();
				type = ffrom[i].getType().getName();
				Object o = ffrom[i].get(from);
				try {
					fto[i].set(to, o);
				}catch(Exception e) {
					Logger.warn("copy fail "+ key + " " + o + " " + type, e);
				}
			}
		}

		// 続いてsuperクラス
		cfrom = cfrom.getSuperclass();
		cto = cto.getSuperclass();

		ffrom =  cfrom.getDeclaredFields();
		fto =  cto.getDeclaredFields();
		for(int i = 0; i < ffrom.length; i++) {
			ffrom[i].setAccessible(true);
			fto[i].setAccessible(true);
			if(Modifier.isTransient(ffrom[i].getModifiers())) {
				continue;
			}
			if (!Modifier.isStatic(ffrom[i].getModifiers())) {
				String key = null,  type = null;
				key = ffrom[i].getName();
				type = ffrom[i].getType().getName();
				Object o = ffrom[i].get(from);

				try {
					fto[i].set(to, o);
				}catch(Exception e) {
					Logger.warn("copy fail "+ key + " " + o + " " + type, e);
				}
			}
		}
		return true;
	}

	/**
	 * クラスを名前から生成する
	 * @param className クラスの完全修飾名
	 * @return オブジェクト
	 * @throws Exception 何らかの例外
	 */
	public static Object createObject(String className) throws Exception {
		return Class.forName(className).newInstance();
	}

	/**
	 * クラスの完全修飾名からパッケージ名を切り出す。(.区切りの最後の.以降を切り捨てる)
	 * @param className クラスの完全修飾名
	 * @return パッケージ名
	 */
	public static String getPackageName(String className) {
		String[] part = className.split("\\.");
		StringBuilder sb = new StringBuilder();
		for(int i = 0; i < part.length-1; i++) {
			sb.append(part[i]);
			if(i < part.length-2) {
				sb.append(".");
			}
		}
		return sb.toString();

	}

}
